# -*- coding: utf-8 -*-

"""Functions for model parsing yaml into a working model instance. Involves a lot of dynamic function definition...
"""

import yaml
import re
import os
import sys
import tempfile
import jax
import jaxlib
import jax.numpy as jnp
import importlib.util as iu
from copy import deepcopy, copy
from inspect import getmembers, isfunction
from jax.experimental.host_callback import id_print as jax_print
from .write_dynamic_functions import *
from .checks import *
from ..utilities import grids, dists, interp

# initialize model cache
cached_mdicts = ()
cached_models = ()


def d2jnp(x): return jnp.array(list(x.values()))


def _load_as_module(path, add_to_path=True):
    """load a file as a module
    """

    if add_to_path:
        directory = os.path.dirname(path)
        sys.path.append(directory)

    modname = os.path.splitext(os.path.basename(path))[0]
    spec = iu.spec_from_file_location(modname, path)
    module = iu.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def parse(mfile):
    """Parse model dictionary from yaml file. This can be desirable if values should be exchanged before loading the model.

    Parameters
    ----------
    mfile : string
        path to a yaml file to be parsed

    Returns
    -------
    mdict : dict
        the parsed yaml as a dictionary
    """

    f = open(mfile)
    mtxt = f.read()
    f.close()

    mtxt = mtxt.replace("^", "**")
    mtxt = re.sub(r"@ ?\n", " ", mtxt)
    # try to detect if `~` wants to be a `-`
    mtxt = mtxt.replace("\n ~ ", "\n - ")
    mtxt = mtxt.replace("\n  ~ ", "\n  - ")
    mtxt = mtxt.replace("   ~ ", "   - ")

    # get dict
    mdict = yaml.safe_load(mtxt)
    # create nice shortcuts
    mdict['path'] = mfile
    mdict["vars"] = mdict["variables"]
    # load file with additional functions as module (if it exists)
    _parse_external_functions_file(mdict)

    return mdict


def _eval_strs(vdict, context={}):
    """Evaluate a dictionary of strings into a given context
    """

    if vdict is None:
        return None, context
    else:
        vdict = vdict.copy()
    context = context.copy()

    for v in vdict:
        if isinstance(vdict[v], str):
            context[v] = eval(vdict[v], context)
            vdict[v] = eval(v, context)
        else:
            context[v] = vdict[v]

    return vdict, context


def _parse_external_functions_file(model):
    """Parse the functions file.
    """
    try:
        # prepare path
        if not os.path.isabs(model["functions_file"]):
            yaml_dir = os.path.dirname(model["path"])
            model["functions_file"] = os.path.join(
                yaml_dir, model["functions_file"])

        # store content
        f = open(model["functions_file"])
        model['functions_file_plain'] = f.read()
        f.close()

    except KeyError:
        pass

    return


def _initialize_context():
    context = globals().copy()
    default_funcs = {'log': jnp.log,
                     'exp': jnp.exp,
                     'sqrt': jnp.sqrt,
                     'max': jnp.maximum,
                     'min': jnp.minimum}

    context.update(default_funcs)
    return context


def _initialize_cache():
    cache = {}
    cache['steady_state'] = ()
    cache['steady_state_keys'] = ()
    return cache


def _load_external_functions_file(model, context):
    """Load the functions file as a module.
    """

    try:
        # load as a module
        module = _load_as_module(model["functions_file"])

        def func_or_compiled(func): return isinstance(
            func, jaxlib.xla_extension.PjitFunction) or isfunction(func)
        for m in getmembers(module, func_or_compiled):
            context[m[0]] = m[1]

    except KeyError:
        pass

    return False


def _compile_init_values(evars, pars, initvals, stst):
    """Combine all available information in initial guesses.
    """

    ufixed_vars = {v: .95 for v in evars if v not in stst}
    ufixed_pars = {v: .95 for v in pars if v not in stst}

    # get inital values to test the function
    init = {**ufixed_vars, **ufixed_pars}

    # structure: aggregate values first, then values of decisions functions
    if initvals is not None:
        for v in init:
            # assign aggregate values
            if v in initvals:
                init[v] = initvals[v]
    return init


def _define_subdict_if_absent(parent, sub):
    try:
        return parent[sub]
    except KeyError:
        parent[sub] = {}
        return parent[sub]


def _define_function(func_str, context):
    """Define functions from string. Writes the function into a temporary file in order to get meaningful debug traces.
    """

    # use a termporary file to get nice debug traces if things go wrong
    tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)
    tmpf.write(func_str)
    tmpf.close()

    # define the function
    exec(compile(open(tmpf.name).read(), tmpf.name, "exec"), context)

    return tmpf.name


def _get_pre_stst_mapping(init_vals, fixed_values, evars, par_names):
    """Define the mapping from init & fixed vals to model variables & parameters
    """
    x2var = jnp.zeros((len(evars), len(init_vals)))
    x2par = jnp.zeros((len(par_names), len(init_vals)))
    fixed2var = jnp.zeros((len(evars), len(fixed_values)))
    fixed2par = jnp.zeros((len(par_names), len(fixed_values)))

    for i, v in enumerate(init_vals):
        if v in evars:
            x2var = x2var.at[evars.index(v), i].set(1)
        else:
            x2par = x2par.at[par_names.index(v), i].set(1)
    for i, v in enumerate(fixed_values):
        if v in evars:
            fixed2var = fixed2var.at[evars.index(v), i].set(1)
        else:
            fixed2par = fixed2par.at[par_names.index(v), i].set(1)

    return x2var, x2par, fixed2var, fixed2par


def compile_stst_inputs(model):

    par_names = model["parameters"]
    evars = model["variables"]
    context = model["context"]
    # remove old values from model context
    [context.pop(key) for key in par_names if key in context]
    [context.pop(key) for key in evars if key in context]

    # collect fixed values and ensure ordering and lenght is fine
    fixed_values = _define_subdict_if_absent(
        model["steady_state"], "fixed_values")
    fixed_values_evaluated, context = _eval_strs(fixed_values, context=context)
    fixed_values_names = tuple(
        sorted([k for k in fixed_values_evaluated if k in par_names or k in evars]))
    fixed_evaluated = {
        k: fixed_values_evaluated[k] for k in fixed_values_names}

    # collect initial guesses
    init_guesses, context = _eval_strs(model["steady_state"].get(
        "init_guesses"), context=context)
    init_vals = _compile_init_values(
        evars, par_names, init_guesses, fixed_evaluated)

    # get the initial decision functions
    if model.get('distributions'):
        dist_names = list(model['distributions'].keys())
        # for now assume that this must be present
        init_wf = jnp.array([init_guesses[dec_input]
                            for dec_input in model['decisions']['inputs']])
        # allow `n` in distributions to be a variable from the context (so far only for one distribution)
        for d in model['distributions'][dist_names[0]].values():
            if isinstance(d['n'], str):
                d['n'] = eval(d['n'], context)
        # check if initial decision functions and the distribution have same shapes
        check_shapes(model['distributions'], init_wf, dist_names)
    else:
        init_wf = jnp.array(None)

    # define func_pre_stst
    mapping = _get_pre_stst_mapping(
        init_vals, fixed_evaluated, evars, par_names)

    return init_vals, fixed_evaluated, init_wf, mapping


def load(
    model_ref,
    raise_errors=True,
    verbose=True
):
    """Load a model from a dictionary or a YAML file.

    Parameters
    ----------
    model_ref : dict or string
        either a dictionary or the path to a YAML file to be parsed
    raise_errors : bool, optional
        whether to raise errors while checking. False will let the model fail siliently for debugging. Defaults to True
    verbose : bool, optional
        inform that parsing is done. Defaults to True

    Returns
    -------
    model : PizzaModel
        The parsed model
    """

    global cached_mdicts, cached_models
    from ..__init__ import PizzaModel

    # parse if this is a path to yaml file
    if isinstance(model_ref, str):
        full_path = model_ref
        model_ref = parse(model_ref)
        model_ref['path'] = full_path

    # define the model dictionary as key for cached models
    mdict = deepcopy(model_ref)
    stst_subdict = mdict.pop('steady_state') if 'steady_state' in mdict else {}
    # check if model is already cached
    if mdict in cached_mdicts:
        model = copy(cached_models[cached_mdicts.index(mdict)])
        # always use the fresh current stst sec from yaml
        model['steady_state'] = stst_subdict
        if verbose:
            print("(load:) Loading cached model.")
        return model

    # make it a model
    model = PizzaModel(model_ref)
    # initialize objects
    model['context'] = _initialize_context()
    model['cache'] = _initialize_cache()
    _load_external_functions_file(model, model['context'])

    # compile globals & definitions
    _ = _define_subdict_if_absent(model, "globals")
    model['context'].update(model['globals'])
    defs = model.get("definitions")
    defs = '' if defs is None else defs
    defs = '\n'.join(defs) if isinstance(defs, list) else defs
    exec(defs, model['context'])
    # get aggregate equations
    eqns = model["equations"].copy()

    # check if there are dublicate variables
    check_dublicates(model["variables"])
    check_dublicates(model.get("parameters"))
    evars = check_determinancy(model["variables"], eqns)
    # check if each variable is defined in time t (only defining xSS does not give a valid root)
    check_if_defined(evars, eqns, model.get('decisions'),
                     model.get('skip_check_if_defined'))

    # create fixed (time invariant) grids
    grids.create_grids(model.get('distributions'), model["context"], verbose)
    shocks = model.get("shocks") or ()

    # initialize storages
    _ = _define_subdict_if_absent(model, "func_strings")
    _ = _define_subdict_if_absent(model, "steady_state")
    par_names = _define_subdict_if_absent(model, "parameters")
    if 'lambda' in evars + par_names:
        raise NameError(
            "Variables or parameters must not use the name of python's build-in functions \"lambda\".")
    if isinstance(par_names, dict):
        raise TypeError(
            f'parameters must be a list and not {type(par_names)}.')

    # get function strings for decisions and distributions, if they exist
    if model.get('decisions'):
        decisions_outputs = model['decisions']['outputs']
        decisions_inputs = model['decisions']['inputs']
        model['func_strings']["func_backw"] = compile_backw_func_str(
            evars, par_names, shocks, decisions_inputs, decisions_outputs, model['decisions']['calls'])
        _define_function(model['func_strings']
                         ['func_backw'], model['context'])
    else:
        decisions_outputs = []
        decisions_inputs = []

    if model.get('distributions'):
        # get names of distributions and the forward functions
        dist_names = list(model['distributions'].keys())
        get_forw_funcs(model)
    else:
        dist_names = []

    # get strings that contains the function definitions
    model['func_strings']["func_eqns"] = compile_eqn_func_str(evars, deepcopy(eqns), par_names, eqns_aux=model.get(
        'aux_equations'), shocks=shocks, distributions=dist_names, decisions_outputs=decisions_outputs)

    # writing to tempfiles helps to get nice debug traces if the model does not work
    _define_function(model['func_strings']["func_eqns"], model['context'])
    # compile fixed and initial values
    stst_inputs = compile_stst_inputs(model)
    # try if function works on initvals
    try:
        check_initial_values(model, shocks, *stst_inputs)
    except:
        if raise_errors:
            raise
    # add new model to cache
    cached_mdicts += (mdict,)
    cached_models += (model,)

    if verbose:
        print("(load:) Parsing done.")

    return model



# -*- coding: utf-8 -*-

import os
import sys
import jax.numpy as jnp
from testbook import testbook
# autopep8: off
sys.path.insert(0, os.path.abspath("."))
import econpizza as ep
# autopep8: on

filepath = os.path.dirname(__file__)

ipynb_path = os.path.join(filepath, '..', '..', 'docs',
                          'tutorial', 'hank2.ipynb')


@testbook(ipynb_path, execute=True, timeout=120)
def test_this(tb):
    assert True



# -*- coding: utf-8 -*-

import os

pth = os.path.dirname(__file__)

nk = os.path.join(pth, "nk.yml")
bh = os.path.join(pth, "bh.yml")
dsge = os.path.join(pth, "dsge.yml")
ghls = os.path.join(pth, "ghls.yml")
hank = os.path.join(pth, "hank_with_comments.yml")
hank_labor = os.path.join(pth, "hank_labor.yml")
hank2 = os.path.join(pth, "hank2.yml")
hank2_no_capital = os.path.join(pth, "hank2_no_capital.yml")
tank = os.path.join(pth, "tank.yml")



# -*- coding: utf-8 -*-

import sys
import time
import jax
import jax.numpy as jnp
from grgrjax import newton_jax_jit, val_and_jacfwd


msgs = (
    ", root finding did not converge",
    ", ftol not reached in root finding",
    ", contains NaNs",
    ", contains infs",
    ", max_iter reached",
)


def find_path_shooting(
    model,
    x0=None,
    shock=None,
    horizon=30,
    init_path=None,
    max_horizon=200,
    max_loops=100,
    max_iter=None,
    tol=1e-5,
    raise_error=False,
    verbose=True,
):
    """Find the expected trajectory given an initial state. A good strategy is to first set `tol` to a low value (e.g. 1e-3) and check for a good max_horizon. Then, set max_horizon to a reasonable value and let max_loops be high.

    NOTE: this is painfully slow since the migration to JAX.

    Parameters
    ----------
    model : PizzaModel
        PizzaModel instance
    x0 : array
        initial state
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`
    horizon : int, optional
        number of periods to simulate
    init_path : array, optional
        a first guess on the trajectory. Normally not necessary
    max_horizon : int, optional
        number of periods until the system is assumed to be back in the steady state. A good idea to set this corresponding to the respective problem. Note that a horizon too far away may cause the accumulation of numerical errors.
    max_loops : int, optional
        number of repetitions to iterate over the whole trajectory. Should eventually be high.
    max_iterations : int, optional
        number of iterations. Default is `max_horizon`. It should not be lower than that (and will raise an error). Normally it should not be higher, better use `max_loops` instead.
    tol : float, optional
        convergence criterion
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent

    Returns
    -------
    x_fin : array
        array of the trajectory
    x_lin : array or None
        array of the trajectory based on the linear model. Will return None if the linear model is unknown
    fin_flag : int
        error code
    """

    st = time.time()

    if max_iter is None:
        max_iter = max_horizon
    elif max_iter < max_horizon:
        Exception(
            "max_iter should be higher or equal max_horizon, but is %s and %s."
            % (max_iter, max_horizon)
        )

    stst = jnp.array(list(model["stst"].values()))
    nvars = len(model["variables"])
    shocks = model.get("shocks") or ()
    pars = jnp.array(list(model["pars"].values()))
    func = jax.jit(model['context']["func_eqns"])

    x_fin = jnp.empty((horizon + 1, nvars))
    x_fin = x_fin.at[0].set(list(x0) if x0 is not None else stst)

    x = jnp.ones((horizon + max_horizon + 1, nvars)) * stst
    x = x.at[0].set(x_fin[0])

    if init_path is not None:
        x = x.at[1: len(init_path)].set(init_path[1:])

    tshock = jnp.zeros(len(shocks))

    fin_flag = jnp.zeros(5, dtype=bool)
    old_clock = time.time()

    @jax.jit
    def solve_current(pars, shock, XLag, XLastGuess, XPrime):
        """Solves for one period.
        """

        # partial_func = jax.tree_util.Partial(func, XLag=XLag, XPrime=XPrime, XSS=stst, shocks=shock, pars=pars)
        def partial_func(x): return func(XLag, x, XPrime, stst, shock, pars)
        jav = val_and_jacfwd(partial_func)
        partial_jav = jax.tree_util.Partial(jav)
        res = newton_jax_jit(partial_jav, XLastGuess, verbose=False)

        return res[0], res[3]

    try:
        for i in range(horizon):

            loop = 1
            cnt = 2
            flag = jnp.zeros_like(fin_flag)

            while True:

                x_old = x[1].copy()
                imax = min(cnt, max_horizon)

                flag_loc = jnp.zeros(2, dtype=bool)

                for t in range(imax):

                    if not t and not i and shock is not None:
                        tshock.at[shocks.index(shock[0])].set(shock[1])
                    else:
                        tshock.at[:].set(0)

                    x_new, flag_ftol = solve_current(
                        pars, tshock, x[t], x[t + 1], x[t + 2])
                    flag_ftol = ~flag_ftol

                    x = x.at[t + 1].set(x_new)

                    flag_loc = flag_loc.at[0].set(flag_loc[0])
                    flag_loc = flag_loc.at[1].set(
                        flag_loc[2] or not flag_ftol)

                flag = flag.at[2].set(flag[2] or jnp.any(jnp.isnan(x)))
                flag = flag.at[3].set(flag[3] or jnp.any(jnp.isinf(x)))

                if cnt == max_iter:
                    if loop < max_loops:
                        loop += 1
                        cnt = 2
                    else:
                        flag = flag.at[4].set(flag[4] or True)

                err = jnp.abs(x_old - x[1]).max()

                clock = time.time()
                if verbose and clock - old_clock > 0.5:
                    old_clock = clock
                    print(
                        "   Period{:>4d} | loop{:>5d} | iter.{:>5d} | flag{:>3d} | error: {:>1.8e}".format(
                            i, loop, cnt, 2 ** jnp.arange(5) @ fin_flag, err
                        )
                    )

                if (err < tol and cnt > 2) or flag.any():
                    flag = flag.at[:2].set(flag[2] or flag_loc)
                    if raise_error and flag.any():
                        mess = [i * bool(j) for i, j in zip(msgs, flag)]
                        raise Exception("Aborting%s" % "".join(mess))
                    fin_flag |= flag
                    break

                cnt += 1

            x_fin = x_fin.at[i + 1].set(x[1])
            x = x[1:].copy()

    except Exception as error:
        try:
            raise type(error)(
                str(error)
                + " (raised in period %s during loop %s for forecast %s steps ahead)"
                % (i, loop, t)
            ).with_traceback(sys.exc_info()[2])
        except AttributeError:
            raise error

    fin_flag = fin_flag.at[1].set(fin_flag[1] and not fin_flag[0])
    mess = [i * bool(j) for i, j in zip(msgs, fin_flag)]
    fin_flag = 2 ** jnp.arange(5) @ fin_flag

    if verbose:
        duration = jnp.round(time.time() - st, 3)
        print("(find_path:) Pizza done after %s seconds%s." %
              (duration, "".join(mess)))

    return x_fin, fin_flag



# -*- coding: utf-8 -*-

import os
import sys
import jax.numpy as jnp
# autopep8: off
sys.path.insert(0, os.path.abspath("."))
import econpizza as ep
# autopep8: on

filepath = os.path.dirname(__file__)


def test_bh(create=False):

    mod = ep.load(ep.examples.bh, raise_errors=False)
    _ = mod.solve_stst()

    state = jnp.zeros(len(mod["variables"]))
    state = state.at[:-1].set([0.1, 0.2, 0.0])

    x, flag = ep.find_path_shooting(
        mod, state, horizon=50, max_horizon=500, tol=1e-8, verbose=2)

    path = os.path.join(filepath, "cache", "bh.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_hank_labor(create=False):

    mod_dict = ep.parse(ep.examples.hank_labor)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-8)

    shocks = ('e_beta', .005)

    x, flag = mod.find_path(shocks, horizon=50)

    path = os.path.join(filepath, "cache", "hank_labor.npy")

    assert flag[0] == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_solid(create=False):

    mod_dict = ep.parse(ep.examples.hank)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-8)

    shocks = ('e_beta', .005)

    x, flag = mod.find_path(shocks, use_solid_solver=True,
                            horizon=20, chunk_size=90)

    path = os.path.join(filepath, "cache", "hank_solid.npy")

    assert flag[0] == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def create_all():

    test_bh(create=True)
    test_hank_labor(create=True)
    test_solid(create=True)



"""interpolation tools
"""

import jax
import jax.numpy as jnp
from functools import partial
from jax._src.typing import Array


interpolate_numpy = jnp.vectorize(jnp.interp, signature='(nq),(n),(n)->(nq)')


@partial(jnp.vectorize, signature='(n),(nq),(n)->(nq)')
def interpolate(x: Array, xq: Array, y: Array) -> Array:
    """Efficient linear interpolation exploiting monotonicity.

    Complexity O(n+nq), so most efficient when x and xq have comparable number of points.
    Extrapolates linearly when xq out of domain of x.

    Parameters
    ----------
    x  : array (n), ascending input points
    xq : array (nq), ascending query points
    y  : array (n), data points

    Returns
    ----------
    yq : array (nq), interpolated points
    """

    nx = x.shape[0]

    xi = jnp.minimum(jnp.searchsorted(x, xq, side='right') - 1, nx-2)
    xqpi_cur = (x[xi + 1] - xq) / (x[xi + 1] - x[xi])
    yq = xqpi_cur * y[xi] + (1 - xqpi_cur) * y[xi + 1]

    return yq


def interpolate_fast(xp: Array, x: Array, fp: Array) -> Array:
    """
    One-dimensional linear interpolation for monotonically increasing sample points.
    Returns the one-dimensional piecewise linear interpolant to a function
    with given discrete data points (`xp`, `fp`), evaluated at `x`.

    The vmap'ed version is much faster. Note: does not extrapolate linearly!
    """
    return jax.vmap(jnp.interp)(jax.lax.broadcast(x, (xp.shape[0],)), xp, fp)


def interpolate_coord_robust_vector(x: Array, xq: Array) -> (Array, Array):
    """Get representation xqi, xqpi of xq interpolated against x:
    xq = xqpi * x[xqi] + (1-xqpi) * x[xqi+1]

    Parameters
    ----------
    x    : array (n), ascending data points
    xq   : array (nq), ascending query points

    Returns
    ----------
    xqi  : array (nq), indices of lower bracketing gridpoints
    xqpi : array (nq), weights on lower bracketing gridpoints
    """

    nx = x.shape[0]

    xqi = jnp.minimum(jnp.searchsorted(x, xq, side='right') - 1, nx-2)
    xqpi = (x[xqi+1] - xq) / (x[xqi+1] - x[xqi])

    return xqi, xqpi


interpolate_coord = jnp.vectorize(
    interpolate_coord_robust_vector, signature='(nq),(nq)->(nq),(nq)')


def interpolate_coord_robust(x: Array, xq: Array, check_increasing=False) -> (Array, Array):
    """Linear interpolation exploiting monotonicity only in data x, not in query points xq.
    Simple binary search, less efficient but more robust.
    xq = xqpi * x[xqi] + (1-xqpi) * x[xqi+1]

    Main application intended to be universally-valid interpolation of policy rules.
    Dimension k is optional.

    Parameters
    ----------
    x    : array (n), ascending data points
    xq   : array (k, nq), query points (in any order)

    Returns
    ----------
    xqi  : array (k, nq), indices of lower bracketing gridpoints
    xqpi : array (k, nq), weights on lower bracketing gridpoints
    """
    if x.ndim != 1:
        raise ValueError(
            'Data input to interpolate_coord_robust must have exactly one dimension')

    if check_increasing and jnp.any(x[:-1] >= x[1:]):
        raise ValueError(
            'Data input to interpolate_coord_robust must be strictly increasing')

    if xq.ndim == 1:
        return interpolate_coord_robust_vector(x, xq)
    else:
        i, pi = interpolate_coord_robust_vector(x, xq.ravel())
        return i.reshape(xq.shape), pi.reshape(xq.shape)


@partial(jnp.vectorize, signature='(nq),(nq),(n)->(nq)')
def apply_coord(x_i: Array, x_pi: Array, y: Array) -> Array:
    """Use representation xqi, xqpi to get yq at xq:
    yq = xqpi * y[xqi] + (1-xqpi) * y[xqi+1]

    Parameters
    ----------
    xqi  : array (nq), indices of lower bracketing gridpoints
    xqpi : array (nq), weights on lower bracketing gridpoints
    y  : array (n), data points

    Returns
    ----------
    yq : array (nq), interpolated points
    """
    return x_pi*y[x_i] + (1-x_pi)*y[x_i+1]


@partial(jnp.vectorize, signature='(ni),(ni,nj)->(nj),(nj)')
def lhs_equals_rhs_interpolate(lhs: Array, rhs: Array) -> (Array, Array):
    """
    Given lhs (i) and rhs (i,j), for each j, find the i such that

        lhs[i] > rhs[i,j] and lhs[i+1] < rhs[i+1,j]

    i.e. where given j, lhs == rhs in between i and i+1.

    Also return the pi such that

        pi*(lhs[i] - rhs[i,j]) + (1-pi)*(lhs[i+1] - rhs[i+1,j]) == 0

    i.e. such that the point at pi*i + (1-pi)*(i+1) satisfies lhs == rhs by linear interpolation.
    """

    rhs_check = rhs.at[-1, :].set(jnp.inf)

    ii = jnp.sum(lhs > rhs_check, 0)
    jj = jnp.arange(lhs.shape[0])

    iout = jnp.maximum(ii - 1, 0)
    err_upper = rhs[ii, jj] - lhs[ii]
    err_lower = rhs[ii - 1, jj] - lhs[ii - 1]
    piout = jnp.where(ii == 0, 1, err_upper / (err_upper - err_lower))

    return iout, piout



# -*- coding: utf-8 -*-

import os
import jax
import time
import jax.numpy as jnp
from grgrjax import val_and_jacrev
from ..parser import d2jnp
from ..parser.build_functions import build_aggr_het_agent_funcs, get_stst_derivatives
from ..parser.checks import check_if_compiled
from ..utilities.jacobian import get_stst_jacobian, get_jac_and_value_sliced
from ..utilities.newton import newton_for_jvp, newton_for_banded_jac, newton_jax_jit_wrapper


def write_cache(model, horizon, pars, stst):
    model['cache']['horizon'] = horizon
    model['cache']['pars'] = pars
    model['cache']['stst'] = stst
    return


def find_path_stacking(
    self,
    shock=None,
    init_state=None,
    init_dist=None,
    pars=None,
    horizon=200,
    use_solid_solver=False,
    skip_jacobian=False,
    verbose=True,
    raise_errors=True,
    **newton_args
):
    """Find the expected trajectory given an initial state.

    Parameters
    ----------
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`
    init_state : array, optional
        initial state, defaults to the steady state values
    init_dist : array, optional
        initial distribution, defaults to the steady state distribution
    pars : dict, optional
        alternative parameters. Warning: do only change those parameters that are invariant to the steady state.
    horizon : int, optional
        number of periods until the system is assumed to be back in the steady state. Defaults to ``200``
    use_solid_solver : bool, optional
        calculate the full jacobian and use a standard Newton method. Defaults to ``False``
    skip_jacobian : bool, optional
        whether to skip the calculation of the steady state sequence space Jacobian. If True, the last cached Jacobian will be used. Defaults to ``False``
    verbose : bool, optional
        degree of verbosity. ``0``/``False`` is silent. Defaults to ``False``
    raise_errors : bool, optional
        whether to raise errors as exceptions, or just inform about them. Defaults to ``True``
    newton_args : optional
        any additional arguments to be passed on to the Newton solver (see the documentations of the solvers)

    Returns
    -------
    x : array
        array of the trajectory
    flag : bool
        Error flag. Returns `False` if the solver was successful, otherwise returns `True`
    """

    st = time.time()
    # only skip jacobian calculation if it exists
    skip_jacobian = skip_jacobian if self['cache'].get(
        'jac_factorized') else False

    # get variables
    stst = d2jnp(self["stst"])
    nvars = len(self["variables"])
    pars = d2jnp(pars if pars is not None else self["pars"])
    shocks = self.get("shocks") or ()

    # get initial guess
    x0 = jnp.array(list(init_state)) if init_state is not None else stst
    dist0 = jnp.array(init_dist) if init_dist is not None else jnp.array(
        self['steady_state'].get('distributions'))
    x_stst = jnp.ones((horizon + 1, nvars)) * stst
    x_init = x_stst.at[0].set(x0)

    # deal with shocks if any
    shock_series = jnp.zeros((horizon-1, len(shocks)))
    if shock is not None:
        try:
            shock_series = shock_series.at[0,
                                           shocks.index(shock[0])].set(shock[1])
        except ValueError:
            raise ValueError(f"Shock '{shock[0]}' is not defined.")

    if not self.get('distributions'):

        if not check_if_compiled(self, horizon, pars, stst) or not self['context'].get('jav_func'):
            # get transition function
            func_eqns = self['context']["func_eqns"]
            jav_func_eqns = val_and_jacrev(func_eqns, (0, 1, 2))
            jav_func_eqns_partial = jax.tree_util.Partial(
                jav_func_eqns, XSS=stst, pars=pars, distributions=[], decisions_outputs=[])
            self['context']['jav_func'] = jav_func_eqns_partial
            # mark as compiled
            write_cache(self, horizon, pars, stst)

        # actual newton iterations
        jav_func_eqns_partial = self['context']['jav_func']
        x_out, f, flag, mess = newton_for_banded_jac(
            jav_func_eqns_partial, nvars, horizon, x_init, shock_series, verbose, **newton_args)

    else:
        if not check_if_compiled(self, horizon, pars, stst) or not self['context'].get('jvp_func'):
            # get derivatives via AD and compile functions
            zero_shocks = jnp.zeros_like(shock_series).T
            build_aggr_het_agent_funcs(self, jnp.zeros_like(
                pars), nvars, stst, zero_shocks, horizon)

            if not use_solid_solver and not skip_jacobian:
                # get steady state partial jacobians
                derivatives = get_stst_derivatives(
                    self, nvars, pars, stst, x_stst, zero_shocks, horizon, verbose)
                # accumulate steady stat jacobian
                get_stst_jacobian(self, derivatives, horizon, nvars, verbose)
            # mark as compiled
            write_cache(self, horizon, pars, stst)

        # get jvp function and steady state jacobian
        jvp_partial = jax.tree_util.Partial(
            self['context']['jvp_func'], x0=x0, dist0=dist0, shocks=shock_series.T, pars=pars)
        if not use_solid_solver:
            jacobian = self['cache']['jac_factorized']
            # actual newton iterations
            x, f, flag, mess = newton_for_jvp(
                jvp_partial, jacobian, x_init, verbose, **newton_args)
        else:
            # define function returning value and jacobian calculated in slices
            value_and_jac_func = get_jac_and_value_sliced(
                (horizon-1)*nvars, jvp_partial, newton_args)
            x, f, flag, mess = newton_jax_jit_wrapper(
                value_and_jac_func, x_init[1:-1].flatten(), **newton_args)
        x_out = x_init.at[1:-1].set(x.reshape((horizon - 1, nvars)))

    # some informative print messages
    duration = time.time() - st
    result = 'done' if not flag else 'FAILED'
    mess = f"(find_path:) Stacking {result} ({duration:1.3f}s). " + mess
    if flag and raise_errors:
        raise Exception(mess)
    elif verbose:
        print(mess)

    return x_out, (flag, f)



import jax
import time
import jax.numpy as jnp
import scipy.sparse as ssp
from jax._src.api import partial
from jax._src.typing import Array
from grgrjax import jax_print


def accumulate(i_and_j: Array, carry: (Array, Array)) -> (Array, int):
    # accumulate effects over different horizons
    jac, horizon = carry
    i = i_and_j // (horizon-2)
    j = i_and_j % (horizon-2)
    jac = jac.at[i+1, :, j+1].add(jac[i, :, j])
    return jac, horizon


@jax.jit
def get_stst_jacobian_jit(derivatives, horizon):
    # load derivatives
    (jac_f2xLag, jac_f2x, jac_f2xPrime), jac_f2do, jac_do2x = derivatives

    # jacobian is quasi-kronecker of f2do x do2x
    jac = jnp.tensordot(jac_f2do[:, ::-1],
                        jac_do2x[..., ::-1, :], jac_f2do.ndim-2)
    jac = jnp.moveaxis(jac, 0, 1)

    # add direct effects of x on f
    jac = jac.at[0, :, 0, :].add(jac_f2x[..., 0])
    jac = jac.at[0, :, 1, :].add(jac_f2xPrime[..., 0])
    jac = jac.at[1, :, 0, :].add(jac_f2xLag[..., 0])

    # accumulate
    jac, _ = jax.lax.fori_loop(0, (horizon-2)**2, accumulate, (jac, horizon))
    return jac


def lu_factor_from_sparse(lu):
    """Translate the output of scipy.sparse.linalg.splu to something that jax.scipy.linalg.lu_solve understands
    """
    pr = lu.perm_r
    pc = lu.perm_c
    n = len(pr)
    pr = jnp.empty(n, dtype=int).at[pr].set(jnp.arange(n))
    lu_factor = lu.L.A - jnp.eye(n) + lu.U.A
    return (lu_factor, pr), pc


def get_stst_jacobian(model, derivatives, horizon, nvars, verbose):
    """Calculate the steady state jacobian
    """
    st = time.time()
    # do the accumulation in jitted jax and flatten
    jac = get_stst_jacobian_jit(derivatives, horizon)
    jac = jac.reshape(((horizon-1)*nvars, (horizon-1)*nvars))
    # store result
    model['cache']['jac'] = jac
    # use sparse SuperLU because it is wayyy faster
    sparse_jac = ssp.csc_matrix(jac)
    sparse_jac_lu = ssp.linalg.splu(sparse_jac)
    model['cache']['jac_factorized'] = lu_factor_from_sparse(sparse_jac_lu)
    if verbose:
        duration = time.time() - st
        print(
            f"(get_jacobian:) Jacobian accumulation and decomposition done ({duration:1.3f}s).")

    return 0


def vmapped_jvp(jvp, primals, tangents):
    """Compact version of jvp_vmap from grgrjax
    """
    pushfwd = partial(jvp, primals)
    y, jac = jax.vmap(pushfwd, out_axes=(None, -1), in_axes=-1)(tangents)
    return y, jac


def jac_slicer(i, carry):
    """Calclulates a chunk of the jacobian
    """
    (_, jac), (x, jvp, zeros_slice, marginal_base, chunk_size) = carry
    # get base slice
    base_slice = jax.lax.dynamic_update_slice(
        zeros_slice, marginal_base, (i*chunk_size, len(x)))
    # calculate slice of the jacobian
    f, jac_slice = vmapped_jvp(jvp, x, base_slice)
    # update jacobian
    jac = jax.lax.dynamic_update_slice(jac, jac_slice, (0, i*chunk_size))
    return (f, jac), (x, jvp, zeros_slice, marginal_base, chunk_size)


def jac_and_value_sliced(jvp, chunk_size, zero_slice, eye_chunk, x):
    """Calculate the value and jacobian at `x` while only evaluating chunks of the full jacobian at times. May be necessary due to memmory requirements.
    """
    x_shape = len(x)
    nloops = jnp.ceil(x_shape/chunk_size).astype(jnp.int64)
    init_vals = x, jnp.zeros((x_shape, x_shape))
    args = x, jvp, zero_slice, eye_chunk, chunk_size
    # in essence a wrapper around a for loop over `jac_slicer`
    (f, jac), _ = jax.lax.fori_loop(0, nloops, jac_slicer, (init_vals, args))
    return f, jac


def get_jac_and_value_sliced(dimx, jvp, newton_args):
    """Get the jac_and_value_sliced function. This is necessary because objects depending on chunk_size must be defined outsite the jitted function
    """
    # get chunk_size from optional dictionary
    if 'chunk_size' in newton_args:
        chunk_size = newton_args['chunk_size']
        newton_args.pop('chunk_size')
    else:
        chunk_size = 100

    # define objects that depend on chunk_size
    zero_slice = jnp.zeros((dimx, chunk_size))
    eye_chunk = jnp.eye(chunk_size)
    return jax.tree_util.Partial(jac_and_value_sliced, jvp, chunk_size, zero_slice, eye_chunk)



# -*- coding: utf-8 -*-
"""functions for the two-asset HANK model. Heavily inspired by https://github.com/shade-econ/sequence-jacobian/#sequence-space-jacobian
"""

import jax.numpy as jnp
from econpizza.utilities.interp import interpolate_coord, apply_coord, interpolate, lhs_equals_rhs_interpolate
from grgrjax import jax_print


def egm_init_Wa(b_grid, a_grid, z_grid, sigma_c):
    Wa = (0.6 + 1.1 * b_grid[:, None] +
          a_grid) ** (-sigma_c) * jnp.ones((z_grid.shape[0], 1, 1))
    return Wa


def egm_init_Wb(b_grid, a_grid, z_grid, sigma_c):
    Wb = (0.5 + b_grid[:, None] + 1.2 *
          a_grid) ** (-sigma_c) * jnp.ones((z_grid.shape[0], 1, 1))
    return Wb


def egm_step(Wa_p, Wb_p, a_grid, b_grid, z_grid, e_grid, kappa_grid, beta, sigma_c, rb, ra, chi0, chi1, chi2, Psi1):

    # === STEP 2: Wb(z, b', a') and Wa(z, b', a') ===
    # (take discounted expectation of tomorrow's value function)
    Wb = beta * Wb_p
    Wa = beta * Wa_p
    W_ratio = Wa / Wb

    # === STEP 3: a'(z, b', a) for UNCONSTRAINED ===
    # for each (z, b', a), linearly interpolate to find a' between gridpoints
    # satisfying optimality condition W_ratio == 1+Psi1
    i, pi = lhs_equals_rhs_interpolate(W_ratio, 1 + Psi1)

    # use same interpolation to get Wb and then c
    a_endo_unc = apply_coord(i, pi, a_grid)
    c_endo_unc = apply_coord(i, pi, Wb) ** (-1/sigma_c)

    # === STEP 4: b'(z, b, a), a'(z, b, a) for UNCONSTRAINED ===
    # solve out budget constraint to get b(z, b', a)
    b_endo = (c_endo_unc + a_endo_unc + addouter(-z_grid, b_grid, -(1 + ra) * a_grid)
              + get_Psi_and_deriv(a_endo_unc, a_grid, ra, chi0, chi1, chi2)[0]) / (1 + rb)

    # interpolate this b' -> b mapping to get b -> b', so we have b'(z, b, a)
    # and also use interpolation to get a'(z, b, a)
    # (interpolate_coord and apply_coord work on last axis, so we need to swap 'b' to the last axis, then back when done)
    i, pi = interpolate_coord(b_endo.swapaxes(1, 2), b_grid)
    a_unc = apply_coord(i, pi, a_endo_unc.swapaxes(1, 2)).swapaxes(1, 2)
    b_unc = apply_coord(i, pi, b_grid).swapaxes(1, 2)

    # === STEP 5: a'(z, kappa, a) for CONSTRAINED ===
    # for each (z, kappa, a), linearly interpolate to find a' between gridpoints
    # satisfying optimality condition W_ratio/(1+kappa) == 1+Psi1, assuming b'=0
    lhs_con = W_ratio[:, 0:1, :] / (1 + kappa_grid[None, :, None])
    i, pi = lhs_equals_rhs_interpolate(lhs_con, 1 + Psi1)

    # use same interpolation to get Wb and then c
    a_endo_con = apply_coord(i, pi, a_grid)
    c_endo_con = ((1 + kappa_grid[None, :, None]) ** (-1/sigma_c)
                  * apply_coord(i, pi, Wb[:, 0:1, :]) ** (-1/sigma_c))

    # === STEP 6: a'(z, b, a) for CONSTRAINED ===
    # solve out budget constraint to get b(z, kappa, a), enforcing b'=0
    b_endo = (c_endo_con + a_endo_con
              + addouter(-z_grid, jnp.full(len(kappa_grid),
                         b_grid[0]), -(1 + ra) * a_grid)
              + get_Psi_and_deriv(a_endo_con, a_grid, ra, chi0, chi1, chi2)[0]) / (1 + rb)

    # interpolate this kappa -> b mapping to get b -> kappa
    # then use the interpolated kappa to get a', so we have a'(z, b, a)
    # (interpolate_y does this in one swoop, but since it works on last
    #  axis, we need to swap kappa to last axis, and then b back to middle when done)
    a_con = interpolate(b_endo.swapaxes(1, 2), b_grid,
                        a_endo_con.swapaxes(1, 2)).swapaxes(1, 2)

    # === STEP 7: obtain policy functions and update derivatives of value function ===
    # combine unconstrained solution and constrained solution, choosing latter
    # when unconstrained goes below minimum b
    b = jnp.where(b_unc <= b_grid[0], b_grid[0], b_unc)
    a = jnp.where(b_unc <= b_grid[0], a_con, a_unc)

    # calculate adjustment cost and its derivative
    Psi, _, Psi2 = get_Psi_and_deriv(a, a_grid, ra, chi0, chi1, chi2)

    # solve out budget constraint to get consumption and marginal utility
    c = addouter(z_grid, (1 + rb) * b_grid, (1 + ra) * a_grid) - Psi - a - b

    uc = c ** (-sigma_c)
    uce = e_grid[:, None, None] * uc

    # update derivatives of value function using envelope conditions
    Wa = (1 + ra - Psi2) * uc
    Wb = (1 + rb) * uc

    return Wa, Wb, a, b, c, uce


def adjustment_costs(a, a_grid, ra, chi0, chi1, chi2):
    chi = get_Psi_and_deriv(a, a_grid, ra, chi0, chi1, chi2)[0]
    return chi


def marginal_cost_grid(a_grid, ra, chi0, chi1, chi2):
    # precompute Psi1(a', a) on grid of (a', a) for steps 3 and 5
    Psi1 = get_Psi_and_deriv(a_grid[:, None],
                             a_grid[None, :], ra, chi0, chi1, chi2)[1]
    return Psi1


def get_Psi_and_deriv(ap, a, ra, chi0, chi1, chi2):
    """Adjustment cost Psi(ap, a) and its derivatives with respect to
    first argument (ap) and second argument (a)"""
    a_with_return = (1 + ra) * a
    a_change = ap - a_with_return
    abs_a_change = jnp.abs(a_change)
    sign_change = jnp.sign(a_change)

    adj_denominator = a_with_return + chi0
    core_factor = (abs_a_change / adj_denominator) ** (chi2 - 1)

    Psi = chi1 / chi2 * abs_a_change * core_factor
    Psi1 = chi1 * sign_change * core_factor
    Psi2 = -(1 + ra) * (Psi1 + (chi2 - 1) * Psi / adj_denominator)
    return Psi, Psi1, Psi2


def addouter(z, b, a):
    """Take outer sum of three arguments: result[i, j, k] = z[i] + b[j] + a[k]"""
    return z[:, None, None] + b[:, None] + a


def income(e_grid, tax, w, N, transfers=0.):
    z_grid = (1 - tax) * w * N * e_grid + transfers
    return z_grid



# -*- coding: utf-8 -*-

import jax
import jax.numpy as jnp


def solve_linear_state_space(
    self,
    raise_error=True,
    check_contraction=False,
    lti_max_iter=1000,
    verbose=True,
):
    """Does half-way SGU, solves the model using linear time iteration and uses Klein's method to check for determinancy if the solution fails"""

    if self.get('distributions'):
        raise Exception(
            "A linear state-space solution for models with distributions is not implemented.")

    try:
        from grgrlib import klein, speed_kills
    except ImportError:
        raise ImportError(
            "'solve_linear_state_space' requires the 'grgrlib' package to be installed.")

    evars = self["variables"]
    func = self['context']["func_eqns"]
    par = jnp.array(list(self["pars"].values()))
    shocks = self.get("shocks") or ()
    stst = jnp.array(list(self["stst"].values()))
    nshc = len(shocks)
    nsts = len(stst)

    xmult = stst.copy()
    xmult = xmult.at[jnp.isclose(stst, 0)].set(1)

    zshock = jnp.zeros(len(shocks))
    fx = func(stst, stst, stst, stst, jnp.zeros(nshc), par)

    AA = jax.jacfwd(lambda err: func(stst, stst, err,
                    stst, zshock, par))(stst) * xmult
    BB = jax.jacfwd(lambda err: func(stst, err, stst,
                    stst, zshock, par))(stst) * xmult
    CC = jax.jacfwd(lambda err: func(err, stst, stst,
                    stst, zshock, par))(stst) * xmult
    DD = jax.jacfwd(lambda err: func(stst, stst, stst, stst, err, par))(zshock)
    DD = DD.reshape((nsts, len(shocks)))

    A = jnp.pad(AA, ((0, nshc), (0, nshc)))
    B = jax.scipy.linalg.block_diag(BB, jnp.eye(nshc))
    C = jnp.block([[CC, DD], [jnp.zeros((nshc, A.shape[1]))]])

    self["ABC"] = A, B, C

    I = jnp.eye(nsts + nshc)
    Z = jnp.zeros_like(I)
    P = jnp.block([[B, A], [I, Z]])
    M = jnp.block([[C, Z], [Z, -I]])

    mess = ""
    success = True

    try:
        try:
            lam = -speed_kills(
                P, M, nsts + nshc, max_iter=lti_max_iter, verbose=verbose - 1
            )[1]

        except:
            _, lam = klein(P, M, nsts + nshc, verbose=verbose - 1)

        self["lin_pol"] = -lam[:nsts, :nsts], -lam[:nsts, nsts:]
        mess = "All eigenvalues are good"

    except Exception as error:
        success = False
        if raise_error:
            raise error
        else:
            mess = str(error).strip()
            if mess[-1] == ".":
                mess = mess[:-1]

    if check_contraction:
        A = jnp.linalg.inv(BB) @ AA
        B = jnp.linalg.inv(BB) @ CC

        Aev = jnp.abs(jnp.linalg.eig(A)[0])
        Aev_err = Aev > 1
        Bev = jnp.abs(jnp.linalg.eig(B)[0])
        Bev_err = Bev > 1
        flag += Aev_err.any() or Bev_err.any()
        mess += ", but " if success else ""
        if Aev_err.any():
            mess += "%s forward looking EV%s larger than unity (%s)" % (
                Aev_err.sum(),
                "s are" if Aev_err.sum() > 1 else " is",
                *Aev[Aev_err],
            )
        if Aev_err.any() and Bev_err.any():
            mess += " and "
        if Bev_err.any():
            mess += "%s backward looking EV%s larger than unity (%s)" % (
                Bev_err.sum(),
                "s are" if Bev_err.sum() > 1 else " is",
                *Bev[Bev_err],
            )

    if mess and verbose:
        print(f"(solve_linear:) {mess}{'' if mess[-1] in '.?!' else '.'}")

    return self["ABC"]


def find_path_linear_state_space(self, init_state=None, shock=None, T=30, verbose=True):
    """Solves the expected trajectory as implied by the linear model.

    Parameters
    ----------
    init_state : array
        initial state
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`. NOTE: Not (yet) implemented.
    horizon : int, optional
        number of periods to simulate
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent

    Returns
    -------
    x : array
        array of the trajectory
    flag : bool
        for consistency. Always returns `True`
    """

    stst = jnp.array(list(self["stst"].values()))
    sel = jnp.isclose(stst, 0)

    x0 = jnp.array(list(init_state)) if init_state is not None else stst
    x0 = x0.at[~sel].set(x0[~sel] / stst[~sel] - 1)

    shocks = self.get("shocks") or ()
    tshock = jnp.zeros(len(shocks))
    if shock is not None:
        tshock[shocks.index(shock[0])] = shock[1]

    x_lin = jnp.empty((T+1, len(stst)))
    x_lin = x_lin.at[0].set(x0)

    for t in range(T):
        x_lin = x_lin.at[t + 1].set(self["lin_pol"][0] @ x_lin[t])

        if not t:
            x_lin = x_lin.at[t + 1].add(self["lin_pol"][1] @ tshock)

    x_lin = x_lin.at[:, ~sel].set(((1 + x_lin) * stst)[:, ~sel])

    return x_lin, True



# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
from copy import deepcopy
from grgrjax import newton_jax, val_and_jacfwd, amax
from ..parser import compile_stst_inputs, d2jnp
from ..parser.build_functions import get_func_stst


def solver(jval, fval):
    """A default solver to solve indetermined problems.
    """
    return jnp.linalg.pinv(jval) @ fval


def _get_stst_dist_objs(self, res, maxit_backwards, maxit_forwards):
    """Get the steady state distribution and decision outputs, which is an auxilliary output of the steady state function. Compile error messages if things go wrong.
    """

    res_backw, res_forw = res['aux']
    wfSS, decisions_output, cnt_backwards = res_backw
    distSS, cnt_forwards = res_forw
    decisions_output_names = self['decisions']['outputs']

    # compile informative message
    mess = ''
    if jnp.isnan(jnp.array(wfSS)).any() or jnp.isnan(jnp.array(decisions_output)).any():
        mess += f"Backward iteration returns NaNs. "
    elif jnp.isnan(distSS).any():
        mess += f"Forward iteration returns NaNs. "
    elif distSS.min() < 0:
        mess += f"Distribution contains negative values ({distSS.min():0.1e}). "
    if cnt_backwards == maxit_backwards:
        mess += f'Maximum of {maxit_backwards} backwards calls reached. '
    if cnt_forwards == maxit_forwards:
        mess += f'Maximum of {maxit_forwards} forwards calls reached. '

    # TODO: this should loop over the objects in distSS/wfSS and store under the name of the distribution/decisions (i.e. 'D' or 'Va')
    self['steady_state']["distributions"] = distSS
    self['steady_state']['value_functions'] = wfSS
    self['steady_state']['decisions'] = {
        oput: decisions_output[i] for i, oput in enumerate(decisions_output_names)}

    return mess


def solve_stst(self, tol=1e-8, maxit=15, tol_backwards=None, maxit_backwards=2000, tol_forwards=None, maxit_forwards=5000, force=False, raise_errors=True, check_rank=False, verbose=True, **newton_kwargs):
    """Solves for the steady state.

    Parameters
    ----------
    tol : float, optional
        tolerance of the Newton method, defaults to ``1e-8``
    maxit : int, optional
        maximum of iterations for the Newton method, defaults to ``15``
    tol_backwards : float, optional
        tolerance required for backward iteration. Defaults to ``tol``
    maxit_backwards : int, optional
        maximum of iterations for the backward iteration. Defaults to ``2000``
    tol_forwards : float, optional
        tolerance required for forward iteration. Defaults to ``tol*1e-2``
    maxit_forwards : int, optional
        maximum of iterations for the forward iteration. Defaults to ``5000``
    force : bool, optional
        force recalculation of steady state, even if it is already evaluated. Defaults to ``False``
    raise_errors : bool, optional
        raise an error if Newton method does not converge. Useful for debuggin models. Defaults to ``True``
    check_rank : bool, optional
        force checking the rank of the Jacobian, even if the Newton method was successful. Defualts to ``False``
    verbose : bool, optional
        level of verbosity. Defaults to ``True``
    newton_kwargs : keyword arguments
        keyword arguments passed on to the Newton method

    Returns
    -------
    rdict : dict
        a dictionary containing information about the root finding result. Note that the results are added to the model (PizzaModel instance) automatically, `rdict` is hence only useful for model debugging.
    """

    st = time.time()
    evars = self["variables"]
    par_names = self["parameters"]
    shocks = self.get("shocks") or ()

    # default setup
    tol_backwards = tol if tol_backwards is None else tol_backwards
    tol_forwards = tol*1e-2 if tol_forwards is None else tol_forwards
    setup = tol, maxit, tol_backwards, maxit_backwards, tol_forwards, maxit_forwards

    # parse and compile steady_state section from yaml
    init_vals, fixed_vals, init_wf, pre_stst_mapping = compile_stst_inputs(
        self)

    # check if model is already cached
    key = str(f'{setup};{d2jnp(fixed_vals)};{d2jnp(init_vals)};{init_wf.sum()}')
    cache = self['cache']
    if key in self['cache']['steady_state_keys'] and not force:
        self['steady_state'] = cache['steady_state'][cache['steady_state_keys'].index(
            key)]
        self["stst"], self["pars"] = deepcopy(
            self['steady_state']["values_and_pars"])
        if verbose:
            print(
                f"(solve_stst:) Steady state already {'known' if self['steady_state']['newton_result']['success'] else 'FAILED'}.")
        return self["steady_state"]["newton_result"]

    # get all necessary functions
    func_eqns = self['context']['func_eqns']
    func_backw = self['context'].get('func_backw')
    func_forw_stst = self['context'].get('func_forw_stst')
    func_pre_stst = self['context']['func_pre_stst']

    # get initial values for heterogenous agents
    decisions_output_init = self['context']['init_run'].get(
        'decisions_output')

    # get the actual steady state function
    func_stst = get_func_stst(func_backw, func_forw_stst, func_eqns, shocks, init_wf, decisions_output_init, fixed_values=d2jnp(
        fixed_vals), pre_stst_mapping=pre_stst_mapping, tol_backw=tol_backwards, maxit_backw=maxit_backwards, tol_forw=tol_forwards, maxit_forw=maxit_forwards)
    # store jitted stst function that returns jacobian and func. value
    self["context"]['func_stst'] = func_stst

    if not self['steady_state'].get('skip'):
        # actual root finding
        res = newton_jax(func_stst, d2jnp(init_vals), maxit, tol,
                         solver=solver, verbose=verbose, **newton_kwargs)
    else:
        f, jac, aux = func_stst(d2jnp(init_vals))
        res = {'x': d2jnp(init_vals),
               'fun': f,
               'jac': jac,
               'success': True,
               'message': 'I blindly took the given values.',
               'aux': aux,
               }

    # exchange those values that are identified via stst_equations
    stst_vals, par_vals = func_pre_stst(
        res['x'], d2jnp(fixed_vals), pre_stst_mapping)
    res['initial_values'] = {'guesses': init_vals, 'fixed': fixed_vals}

    self["stst"] = dict(zip(evars, stst_vals))
    self["pars"] = dict(zip(par_names, par_vals))
    self['steady_state']["newton_result"] = res
    self['steady_state']["values_and_pars"] = deepcopy(
        self["stst"]), deepcopy(self["pars"])

    # calculate dist objects and compile message
    mess = _get_stst_dist_objs(self, res, maxit_backwards,
                               maxit_forwards) if self.get('distributions') else ''
    # calculate error
    err, errarg = amax(res['fun'], True)

    if err > tol and self['steady_state'].get('skip'):
        mess += f"They do not satisfy the required tolerance."
    elif err > tol or not res['success'] or check_rank:
        rank = jnp.linalg.matrix_rank(res['jac'])
        if rank:
            nvars = len(evars)+len(par_names)
            nfixed = len(fixed_vals)
            if rank != nvars - nfixed:
                mess += f"Jacobian has rank {rank} for {nvars - nfixed} degrees of freedom ({nvars} variables/parameters, {nfixed} fixed). "

    # check if any of the fixed variables are neither a parameter nor variable
    if mess:
        not_var_nor_par = list(
            set(self['steady_state']['fixed_values']) - set(evars) - set(par_names))
        mess += f"Fixed value(s) ``{'``, ``'.join(not_var_nor_par)}`` not declared. " if not_var_nor_par else ''

    if err > tol or not res['success']:
        if not res["success"] or raise_errors:
            location = '' if jnp.isnan(
                err) else f" (max. error is {err:1.2e} in eqn. {errarg})"
            mess = f"Steady state FAILED{location}. {res['message']} {mess}"
        else:
            mess = f"{res['message']} WARNING: Steady state error is {err:1.2e} in eqn. {errarg}. {mess}"
        if raise_errors:
            raise Exception(mess)
    elif verbose:
        duration = time.time() - st
        mess = f"Steady state found ({duration:1.5g}s). {res['message']}" + (
            ' WARNING: ' + mess if mess else '')

    # cache everything if search was successful
    self['cache']['steady_state'] += self['steady_state'],
    self['cache']['steady_state_keys'] += key,

    if mess:
        print(f"(solve_stst:) {mess}")

    return res



"""Functions that write other functions.
"""
import re
import jax
from .build_functions import func_forw_generic, func_forw_stst_generic


def compile_func_basics_str(evars, par, shocks):

    func_str = f"""
        \n ({"".join(v + "Lag, " for v in evars)}) = XLag
        \n ({"".join(v + ", " for v in evars)}) = X
        \n ({"".join(v + "Prime, " for v in evars)}) = XPrime
        \n ({"".join(v + "SS, " for v in evars)}) = XSS
        \n ({"".join(p + ", " for p in par)}) = pars
        \n ({"".join(s + ", " for s in shocks)}) = shocks"""

    return func_str


def compile_backw_func_str(evars, par, shocks, inputs, outputs, calls):
    """Compile all information to a string that defines the backward function for 'decisions'.
    """

    if isinstance(calls, str):
        calls = calls.splitlines()

    func_str = f"""def func_backw(XLag, X, XPrime, XSS, WFPrime, shocks, pars):
            {compile_func_basics_str(evars, par, shocks)}
            \n ({"".join(v + ", " for v in inputs)}) = WFPrime
            \n %s
            \n return jnp.array(({"".join(v[:-5] + ", " for v in inputs)})), jnp.array(({", ".join(v for v in outputs)}))
            """ % '\n '.join(calls)

    return func_str


def get_forw_funcs(model):

    distributions = model['distributions']

    if len(distributions) > 1:
        raise NotImplementedError(
            'More than one distribution is not yet implemented.')

    # already prepare for more than one distributions
    for dist_name in distributions:

        dist = distributions[dist_name]

        implemented_endo = (
            'exogenous', 'exogenous_rouwenhorst', 'exogenous_generic')
        implemented_exo = ('endogenous', 'endogenous_log',
                           'endogenous_generic')
        exog = [v for v in dist if dist[v]['type'] in implemented_endo]
        endo = [v for v in dist if dist[v]['type'] in implemented_exo]
        other = [dist[v]['type'] for v in dist if dist[v]
                 ['type'] not in implemented_endo + implemented_exo]

        if len(exog) > 1:
            raise NotImplementedError(
                'Exogenous distributions with more than one dimension are not yet implemented.')
        if len(endo) > 2:
            raise NotImplementedError(
                'Endogenous distributions with more than two dimension are not yet implemented.')
        if other:
            raise NotImplementedError(
                f"Grid(s) of type {str(*other)} not implemented.")

        transition = model['context'][dist[exog[0]]['transition_name']]
        grids = [model['context'][dist[i]['grid_name']] for i in endo]
        indices = [model['decisions']['outputs'].index(i) for i in endo]

        model['context']['func_forw'] = jax.tree_util.Partial(
            func_forw_generic, grids=grids, transition=transition, indices=indices)
        model['context']['func_forw_stst'] = jax.tree_util.Partial(
            func_forw_stst_generic, grids=grids, transition=transition, indices=indices)

    return


def compile_eqn_func_str(evars, eqns, par, eqns_aux, shocks, distributions, decisions_outputs):
    """Compile all information from 'equations' section' to a string that defines the function.
    """

    # start compiling root_container
    for i, eqn in enumerate(eqns):
        eqsplit = re.split("(?<!=)=(?!=)", eqn)
        if len(eqsplit) == 1:
            eqns[i] = f"root_container{i} = {eqn}"
        elif len(eqsplit) == 2:
            eqns[i] = f"root_container{i} = {eqsplit[0]} - ({eqsplit[1]})"
        else:
            raise SyntaxError(f'More than one " = " in equation {i}: "{eqn}"')

    if isinstance(eqns_aux, str):
        eqns_aux = eqns_aux.splitlines()

    eqns_aux_stack = "\n ".join(eqns_aux) if eqns_aux else ""
    eqns_stack = "\n ".join(eqns)

    # compile the final function string
    func_str = f"""def func_eqns(XLag, X, XPrime, XSS, shocks, pars, distributions=[], decisions_outputs=[]):
        {compile_func_basics_str(evars, par, shocks)}
        \n ({"".join(d+', ' for d in distributions)}) = distributions
        \n ({"".join(d+', ' for d in decisions_outputs)}) = decisions_outputs
        \n {eqns_aux_stack}
        \n {eqns_stack}
        \n {"return jnp.array([" + ", ".join("root_container"+str(i) for i in range(len(evars))) + "]).T.ravel()"}"""

    return func_str



# -*- coding: utf-8 -*-

import logging
import os
import jax
import jax.numpy as jnp
from copy import deepcopy

from . import examples
from .__version__ import __version__
from .solvers.steady_state import solve_stst
from .solvers.stacking import find_path_stacking
from .solvers.solve_linear import find_path_linear
from .solvers.solve_linear_state_space import solve_linear_state_space, find_path_linear_state_space
from .solvers.shooting import find_path_shooting
from .parser import parse, load


# set number of cores for XLA
os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={os.cpu_count()}"

jax.config.update("jax_enable_x64", True)

# create local alias
copy = deepcopy


class PizzaModel(dict):
    """Base class for models. Contains all necessary methods and informations.
    """

    def __init__(self, mdict, *args, **kwargs):
        # do not overwrite original input
        mdict = deepcopy(mdict)
        super(PizzaModel, self).__init__(mdict, *args, **kwargs)
        self.__dict__ = self

    def get_distributions(self, trajectory, init_dist=None, shock=None, pars=None):
        """Get all disaggregated variables for a given trajectory of aggregate variables.

        Note that the output objects do, other than the result from `find_path` with stacking, not include the time-T objects and that the given distribution is as from the beginning of each period.

        Parameters
        ----------

        trajectory : array
            a _full_ trajectory of aggregate variables
        init_dist : array, optional
            the initial distribution. Defaults to the steady state distribution
        shock : array, optional
            shock in period 0 as in `(shock_name_as_str, shock_size)`. Defaults to no shock

        Returns
        -------
        rdict : dict
            a dictionary of the distributions
        """

        dist0 = jnp.array(init_dist) if init_dist is not None else jnp.array(
            self['steady_state'].get('distributions'))
        pars = jnp.array(list(self['pars'].values())
                         ) if pars is None else pars
        shocks = self.get("shocks") or ()
        dist_names = list(self['distributions'].keys())
        decisions_outputs = self['decisions']['outputs']
        x = trajectory[1:-1].flatten()
        x0 = trajectory[0]

        # deal with shocks if any
        shock_series = jnp.zeros((len(trajectory)-2, len(shocks)))
        if shock is not None:
            shock_series = shock_series.at[0,
                                           shocks.index(shock[0])].set(shock[1])

        # get functions and execute
        backwards_sweep = self['context']['backwards_sweep']
        forwards_sweep = self['context']['forwards_sweep']
        decisions_output_storage = backwards_sweep(x, x0, shock_series.T, pars)
        dists_storage = forwards_sweep(decisions_output_storage, dist0)

        # store this
        rdict = {oput: decisions_output_storage[i]
                 for i, oput in enumerate(decisions_outputs)}
        rdict.update({oput: dists_storage[i]
                     for i, oput in enumerate(dist_names)})

        return rdict

    solve_stst = solve_stst
    find_path = find_path_stacking
    find_path_linear = find_path_linear
    find_path_shooting = find_path_shooting
    find_path_linear_state_space = find_path_linear_state_space
    solve_linear_state_space = solve_linear_state_space


logging.basicConfig(level=logging.WARNING)




"""Grids and Markov chains"""

import jax
import jax.numpy as jnp
from .dists import stationary_distribution


def log_grid(amax, n, amin=0):
    """Create grid between amin and amax that is equidistant in logs."""
    pivot = jnp.abs(amin) + 0.25
    a_grid = jnp.geomspace(amin + pivot, amax + pivot, n) - pivot
    a_grid = a_grid.at[0].set(amin)  # make sure *exactly* equal to amin
    return a_grid


def mean(x, pi):
    """Mean of discretized random variable with support x and probability mass function pi."""
    return jnp.sum(pi * x)


def variance(x, pi):
    """Variance of discretized random variable with support x and probability mass function pi."""
    return jnp.sum(pi * (x - jnp.sum(pi * x)) ** 2)


def markov_rouwenhorst(rho, sigma, N):
    """Rouwenhorst method analog to markov_tauchen"""

    # Explicitly typecast N as an integer, since when the grid constructor functions
    # (e.g. the function that makes a_grid) are implemented as blocks, they interpret the integer-valued calibration
    # as a float.
    N = int(N)

    # parametrize Rouwenhorst for n=2
    p = (1 + rho) / 2
    Pi = jnp.array([[p, 1 - p], [1 - p, p]])

    # implement recursion to build from n=3 to n=N
    for n in range(3, N + 1):
        P1, P2, P3, P4 = (jnp.zeros((n, n)) for _ in range(4))
        P1 = P1.at[:-1, :-1].set(p * Pi)
        P2 = P2.at[:-1, 1:].set((1 - p) * Pi)
        P3 = P3.at[1:, :-1].set((1 - p) * Pi)
        P4 = P4.at[1:, 1:].set(p * Pi)
        Pi = P1 + P2 + P3 + P4
        Pi = Pi.at[1:-1].divide(2)

    # invariant distribution and scaling
    pi = stationary_distribution(jnp.array(Pi.T))
    y = rouwenhorst_grid_from_stationary(sigma, pi)
    return y, pi, Pi


def rouwenhorst_grid_from_stationary(sigma, stationary_distribution):
    s = jnp.linspace(-1, 1, len(stationary_distribution))
    s *= (sigma / jnp.sqrt(variance(s, stationary_distribution)))
    return jnp.exp(s) / jnp.sum(stationary_distribution * jnp.exp(s))


def create_grids(distributions, context, verbose):
    """Get the strings of functions that define the grids.
    """

    if not distributions:
        return

    grid_strings = ()

    if verbose:
        print("Creating grid variables:")

    # create strings of the function that define the grids
    for dist_name, dist in distributions.items():
        for grid_name, g in dist.items():

            # shortcuts
            g['type'] = 'endogenous_log' if g['type'] == 'endogenous' else g['type']
            g['type'] = 'exogenous_rouwenhorst' if g['type'] == 'exogenous' else g['type']

            g['grid_name'] = grid_name + '_grid'
            # generic grid added by the user
            if g['type'] == 'exogenous_generic':
                g['transition_name'] = grid_name + '_transition'
                if verbose:
                    print(
                        f"    ...expecting the objects '{g['transition_name']}' and '{g['grid_name']}' for the exogenous grid of '{grid_name}'.")
            if g['type'] == 'endogenous_generic':
                if verbose:
                    print(
                        f"    ...expecting the object '{g['grid_name']}' for the endogenous grid of '{grid_name}'.")

            # predefined grids
            if g['type'] == 'exogenous_rouwenhorst':
                grid_strings += f"{grid_name}_grid, {grid_name}_stationary, {grid_name}_transition = grids.markov_rouwenhorst(rho={g['rho']}, sigma={g['sigma']}, N={g['n']})",
                g['transition_name'] = grid_name + '_transition'
                if verbose:
                    print(
                        f"    ...adding exogenous Rouwenhorst-grid for '{grid_name}' with objects '{g['grid_name']}', '{g['transition_name']}' and '{grid_name}_stationary'.")
            if g['type'] == 'endogenous_log':
                grid_strings += f"{g['grid_name']} = grids.log_grid(amin={g['min']}, amax={g['max']}, n={g['n']})",
                if verbose:
                    print(
                        f"    ...adding endogenous log-grid for '{grid_name}' named '{g['grid_name']}'.")

    # execute all of them
    for grid_str in grid_strings:
        exec(grid_str, context)

    return



"""Tools for dealing with distributions
"""

import jax
import jax.numpy as jnp
from jax._src.typing import Array


def expect_transition(Pi: Array, X: Array) -> Array:
    """If Pi is a matrix, multiply Pi times the ith dimension of X and return"""

    shape = X.shape
    # iterate forward using Pi
    X = Pi @ X.reshape((shape[0], -1))
    # reverse steps
    X = X.reshape((Pi.shape[0], *shape[1:]))

    return X


def forward_policy_1d(D: Array, x_i: Array, x_pi: Array) -> Array:

    nZ, _ = D.shape
    Dnew = jnp.zeros_like(D)

    j = jnp.arange(nZ)[:, None]

    Dnew = Dnew.at[j, x_i].add(D * x_pi)
    Dnew = Dnew.at[j, x_i+1].add(D * (1 - x_pi))

    return Dnew


def cond_func(carry):
    (dist, cnt, dist_old), (tol, maxit), _ = carry
    cond0 = jnp.abs(dist-dist_old).max() > tol
    cond1 = cnt < maxit
    return jnp.logical_and(cond0, cond1)


def _body_func_1d(carry):
    (dist, cnt, _), cond_vars, exo_endo = carry
    exog_probs, endog_inds, endog_probs = exo_endo
    dist_new = exog_probs.T @ forward_policy_1d(dist, endog_inds, endog_probs)
    return (dist_new, cnt + 1, dist), cond_vars, exo_endo


def stationary_distribution_forward_policy_1d(endog_inds: Array, endog_probs: Array, exog_probs: Array, tol=1e-10, maxit=1000) -> (Array, int):

    dist = jnp.ones_like(endog_inds, dtype=jnp.float64)
    dist /= dist.sum()

    (dist, cnt, _), _, _ = jax.lax.while_loop(
        cond_func, _body_func_1d, ((dist, 0, dist+1), (tol, maxit), (exog_probs, endog_inds, endog_probs)))
    return dist, cnt


def forward_policy_2d(D: Array, x_i: Array, y_i: Array, x_pi: Array, y_pi: Array) -> Array:

    nZ, _, _ = D.shape
    Dnew = jnp.zeros_like(D)

    j = jnp.arange(nZ)[:, None, None]

    Dnew = Dnew.at[j, x_i, y_i].add(y_pi * x_pi * D)
    Dnew = Dnew.at[j, x_i+1, y_i].add(y_pi * (1 - x_pi) * D)
    Dnew = Dnew.at[j, x_i, y_i+1].add((1 - y_pi) * x_pi * D)
    Dnew = Dnew.at[j, x_i+1, y_i+1].add((1 - y_pi) * (1 - x_pi) * D)

    return Dnew


def _body_func_2d(carry):
    (dist, cnt, _), cond_vars, exo_endo = carry
    exog_probs, endog_inds0, endog_inds1, endog_probs0, endog_probs1 = exo_endo
    pre_exo_dist = forward_policy_2d(
        dist, endog_inds0, endog_inds1, endog_probs0, endog_probs1)
    new_dist = expect_transition(exog_probs.T, pre_exo_dist)
    return (new_dist, cnt + 1, dist), cond_vars, exo_endo


def stationary_distribution_forward_policy_2d(endog_inds0: Array, endog_inds1: Array, endog_probs0: Array, endog_probs1: Array, exog_probs: Array, tol=1e-10, maxit=1000) -> (Array, int):
    # TODO: can be merged with stationary_distribution_forward_policy_1d

    dist = jnp.ones_like(endog_inds0, dtype=jnp.float64)
    dist /= dist.sum()

    exo_endo = exog_probs, endog_inds0, endog_inds1, endog_probs0, endog_probs1
    (dist, cnt, _), _, _ = jax.lax.while_loop(cond_func,
                                              _body_func_2d, ((dist, 0, dist+1), (tol, maxit), exo_endo))

    return dist, cnt


def stationary_distribution(T):
    """Find invariant distribution of a Markov chain by unit eigenvector.
    NOTE: jax has no autodiff support for eig. (there is a version with custom_jvp in grgrwip)
    """

    v, w = jnp.linalg.eig(T)

    # using sorted args instead of np.isclose is neccessary for jax-jitting
    args = jnp.argsort(v)
    unit_ev = w[:, args[-1]]

    return unit_ev.real / unit_ev.real.sum()
